{% extends "base.html" %}

{% block announce %}
  ✨ New project! Read the <a href="https://quansight.com/post/unveiling-ragna-an-open-source-rag-based-ai-orchestration-framework-designed-to-scale-from-research-to-production/"><b>launch announcement blog post.</b></a> ✨
{% endblock %}
